@@ -15,9 +15,14 @@ from .table import table_layout, table_wrapper_width
 
 
 def block_level_layout(context, box, bottom_space, skip_stack,
-                       containing_block, page_is_empty, absolute_boxes,
-                       fixed_boxes, adjoining_margins, discard, max_lines):
+                       containing_block, page_is_empty=True,
+                       absolute_boxes=None, fixed_boxes=None,
+                       adjoining_margins=None, discard=False, max_lines=None):
     """Lay out the block-level ``box``."""
+    absolute_boxes = [] if absolute_boxes is None else absolute_boxes
+    fixed_boxes = [] if fixed_boxes is None else fixed_boxes
+    adjoining_margins = [] if adjoining_margins is None else adjoining_margins
+
     if not isinstance(box, boxes.TableBox):
         resolve_percentages(box, containing_block)
 
@@ -27,6 +32,8 @@ def block_level_layout(context, box, bottom_space, skip_stack,
             box.margin_bottom = 0
 
         if context.current_page > 1 and page_is_empty:
+            # When an unforced break occurs before or after a block-level box,
+            # any margins adjoining the break are truncated to zero.
             # TODO: this condition is wrong, it only works for blocks whose
             # parent breaks collapsing margins. It should work for blocks whose
             # one of the ancestors breaks collapsing margins.
@@ -97,6 +104,8 @@ def block_box_layout(context, box, bottom_space, skip_stack,
                 new_box.margin_bottom + new_box.padding_bottom +
                 new_box.border_bottom_width)
             if columns_bottom_space:
+                remove_placeholders(
+                    context, [new_box], absolute_boxes, fixed_boxes)
                 bottom_space += columns_bottom_space
                 result = columns_layout(
                     context, box, bottom_space, skip_stack,
@@ -114,7 +123,7 @@ def block_box_layout(context, box, bottom_space, skip_stack,
     new_box = result[0]
     if new_box and new_box.is_table_wrapper:
         # Don't collide with floats
-        # http://www.w3.org/TR/CSS21/visuren.html#floats
+        # https://www.w3.org/TR/CSS21/visuren.html#floats
         position_x, position_y, _ = avoid_collisions(
             context, new_box, containing_block, outer=False)
         new_box.translate(
@@ -134,7 +143,7 @@ def block_level_width(box, containing_block):
         # TODO: what is the real text direction?
         direction = 'ltr'
 
-    # http://www.w3.org/TR/CSS21/visudet.html#blockwidth
+    # https://www.w3.org/TR/CSS21/visudet.html#blockwidth
 
     # These names are waaay too long
     margin_l = box.margin_left
@@ -176,7 +185,7 @@ def block_level_width(box, containing_block):
         width = box.width = cb_width - (
             paddings_plus_borders + margin_l + margin_r)
     margin_sum = cb_width - paddings_plus_borders - width
-    if margin_l == 'auto' and margin_r == 'auto':
+    if margin_l == margin_r == 'auto':
         box.margin_left = margin_sum / 2
         box.margin_right = margin_sum / 2
     elif margin_l == 'auto' and margin_r != 'auto':
@@ -221,11 +230,12 @@ def _out_of_flow_layout(context, box, index, child, new_children,
                         adjoining_margins, bottom_space):
     stop = False
     resume_at = None
+    new_child = None
     out_of_flow_resume_at = None
 
     child.position_y += collapse_margin(adjoining_margins)
     if child.is_absolutely_positioned():
-        placeholder = AbsolutePlaceholder(child)
+        new_child = placeholder = AbsolutePlaceholder(child)
         placeholder.index = index
         new_children.append(placeholder)
         if child.style['position'] == 'absolute':
@@ -257,7 +267,7 @@ def _out_of_flow_layout(context, box, index, child, new_children,
         page = context.current_page
         context.running_elements[running_name][page].append(child)
 
-    return stop, resume_at, out_of_flow_resume_at
+    return stop, resume_at, new_child, out_of_flow_resume_at
 
 
 def _break_line(context, box, line, new_children, lines_iterator,
@@ -340,7 +350,7 @@ def _linebox_layout(context, box, index, child, new_children, page_is_empty,
             break
 
         # TODO: this is incomplete.
-        # See http://dev.w3.org/csswg/css3-page/#allowed-pg-brk
+        # See https://drafts.csswg.org/css-page-3/#allowed-pg-brk
         # "When an unforced page break occurs here, both the adjoining
         #  ‘margin-top’ and ‘margin-bottom’ are set to zero."
         # See https://github.com/Kozea/WeasyPrint/issues/115
@@ -423,15 +433,21 @@ def _in_flow_layout(context, box, index, child, new_children, page_is_empty,
 
     if not new_containing_block.is_table_wrapper:
         resolve_percentages(child, new_containing_block)
-        if (child.is_in_normal_flow() and last_in_flow_child is None and
-                collapsing_with_children):
+        if last_in_flow_child is None and collapsing_with_children:
             # TODO: add the adjoining descendants' margin top to
             # [child.margin_top]
             old_collapsed_margin = collapse_margin(adjoining_margins)
-            if child.margin_top == 'auto':
+            # TODO: the margin-top value is set afterwards in
+            # block_level_layout, we shouldn’t duplicate this code
+            child_margin_top = child.margin_top
+            if child_margin_top == 'auto':
                 child_margin_top = 0
-            else:
-                child_margin_top = child.margin_top
+            elif context.current_page > 1 and page_is_empty:
+                if box.style['margin_break'] == 'discard':
+                    child_margin_top = 0
+                elif box.style['margin_break'] == 'auto':
+                    if not context.forced_break:
+                        child_margin_top = 0
             new_collapsed_margin = collapse_margin(
                 adjoining_margins + [child_margin_top])
             collapsed_margin_difference = (
@@ -625,7 +641,7 @@ def block_container_layout(context, box, bottom_space, skip_stack,
     new_children = []
     next_page = {'break': 'any', 'page': None}
     all_footnotes = []
-    broken_out_of_flow = []
+    broken_out_of_flow = {}
 
     last_in_flow_child = None
 
@@ -645,12 +661,14 @@ def block_container_layout(context, box, bottom_space, skip_stack,
 
         if not child.is_in_normal_flow():
             abort = False
-            stop, resume_at, out_of_flow_resume_at = _out_of_flow_layout(
-                context, box, index, child, new_children, page_is_empty,
-                absolute_boxes, fixed_boxes, adjoining_margins,
-                bottom_space)
+            stop, resume_at, new_child, out_of_flow_resume_at = (
+                _out_of_flow_layout(
+                    context, box, index, child, new_children, page_is_empty,
+                    absolute_boxes, fixed_boxes, adjoining_margins,
+                    bottom_space))
             if out_of_flow_resume_at:
-                broken_out_of_flow.append((child, box, out_of_flow_resume_at))
+                broken_out_of_flow[new_child] = (
+                    child, box, out_of_flow_resume_at)
 
         elif isinstance(child, boxes.LineBox):
             (abort, stop, resume_at, position_y,
@@ -692,6 +710,8 @@ def block_container_layout(context, box, bottom_space, skip_stack,
 
         if abort:
             page = child.page_values()[0]
+            remove_placeholders(
+                context, box.children[skip:], absolute_boxes, fixed_boxes)
             for footnote in new_footnotes:
                 context.unlayout_footnote(footnote)
             return (
@@ -716,7 +736,8 @@ def block_container_layout(context, box, bottom_space, skip_stack,
         return (
             None, None, {'break': 'any', 'page': None}, [], False, max_lines)
 
-    context.broken_out_of_flow.extend(broken_out_of_flow)
+    for key, value in broken_out_of_flow.items():
+        context.broken_out_of_flow[key] = value
 
     if collapsing_with_children:
         box.position_y += (
@@ -729,9 +750,9 @@ def block_container_layout(context, box, bottom_space, skip_stack,
         # Top and bottom margins of this box
         if (box.height in ('auto', 0) and
             get_clearance(context, box, collapsed_margin) is None and
-            all(value == 0 for value in [
+            all(value == 0 for value in (
                 box.min_height, box.border_top_width, box.padding_top,
-                box.border_bottom_width, box.padding_bottom])):
+                box.border_bottom_width, box.padding_bottom))):
             collapsing_through = True
         else:
             position_y += collapsed_margin
@@ -761,7 +782,7 @@ def block_container_layout(context, box, bottom_space, skip_stack,
         start=not is_start, end=box_is_fragmented and not discard)
 
     # TODO: See corner cases in
-    # http://www.w3.org/TR/CSS21/visudet.html#normal-block
+    # https://www.w3.org/TR/CSS21/visudet.html#normal-block
     # TODO: See float.float_layout
     if new_box.height == 'auto':
         if context.excluded_shapes and new_box.style['overflow'] != 'visible':
@@ -792,13 +813,15 @@ def block_container_layout(context, box, bottom_space, skip_stack,
     elif bottom_space > -inf and not new_box.is_column:
         # Make the box fill the blank space at the bottom of the page
         # https://www.w3.org/TR/css-break-3/#box-splitting
-        new_box.height = (
+        new_box_height = (
             context.page_bottom - bottom_space - new_box.position_y -
             (new_box.margin_height() - new_box.height))
-        if draw_bottom_decoration:
-            new_box.height += (
-                box.padding_bottom + box.border_bottom_width +
-                box.margin_bottom)
+        if new_box_height > new_box.height:
+            new_box.height = new_box_height
+            if draw_bottom_decoration:
+                new_box.height += (
+                    box.padding_bottom + box.border_bottom_width +
+                    box.margin_bottom)
 
     if next_page['page'] is None:
         next_page['page'] = new_box.page_values()[1]
@@ -820,7 +843,7 @@ def collapse_margin(adjoining_margins):
 def establishes_formatting_context(box):
     """Return whether a box establishes a block formatting context.
 
-    See http://www.w3.org/TR/CSS2/visuren.html#block-formatting
+    See https://www.w3.org/TR/CSS2/visuren.html#block-formatting
 
     """
     return (
@@ -1001,18 +1024,21 @@ def remove_placeholders(context, box_list, absolute_boxes, fixed_boxes):
     For boxes that have been removed in find_earlier_page_break(), remove the
     matching placeholders in absolute_boxes and fixed_boxes.
 
+    Also takes care of removed footnotes and floats.
+
     """
     for box in box_list:
         if isinstance(box, boxes.ParentBox):
             remove_placeholders(
                 context, box.children, absolute_boxes, fixed_boxes)
         if box.style['position'] == 'absolute' and box in absolute_boxes:
-            # box is not in absolute_boxes if its parent has position: relative
             absolute_boxes.remove(box)
-        elif box.style['position'] == 'fixed':
+        elif box.style['position'] == 'fixed' and box in fixed_boxes:
             fixed_boxes.remove(box)
         if box.footnote:
             context.unlayout_footnote(box.footnote)
+        if box in context.broken_out_of_flow:
+            context.broken_out_of_flow.pop(box)
 
 
 def avoid_page_break(page_break, context):
